#pragma once

#include "includes/Definitions.hpp"

namespace SPHERE
{
  struct State
  {
    double Diameter;
    complex128 Index;
    double nMedium;
    State(){}
    State(double &Diameter, complex128 &Index, double &nMedium)
                : Diameter(Diameter), Index(Index), nMedium(nMedium){}

    State(double &Diameter, complex128 &&Index, double &nMedium)
                : Diameter(Diameter), Index(Index), nMedium(nMedium){}

    void ApplyMedium()
    {
      this->Index    /= this->nMedium;
      this->Diameter *= this->nMedium;
    }
  };


  class Set
  {
    public:
      DVector                              Diameter;
      CVector                              Index;
      std::vector<CVector>                 Material;
      DVector                              nMedium;
      bool                                 BoundedIndex;
      std::vector<State>                   States;

      Set(){}
      Set(DVector &Diameter, std::vector<std::vector<complex128>> &Material, DVector &nMedium)
          : Diameter(Diameter), Material(Material), nMedium(nMedium)
          {BoundedIndex=true;}

      Set(DVector &Diameter, std::vector<complex128> &Index, DVector &nMedium)
          : Diameter(Diameter), Index(Index), nMedium(nMedium)
          {BoundedIndex=false;}

      State operator[](size_t idx){return this->States[idx];}

   };
}








namespace CYLINDER
{

  struct State
  {
    double     Diameter;
    complex128 Index;
    double     nMedium;
    State(){}
    State(double &Diameter, complex128 &Index, double &nMedium)
                : Diameter(Diameter), Index(Index), nMedium(nMedium){}

    void ApplyMedium()
    {
      this->Index    /= this->nMedium;
      this->Diameter *= this->nMedium;
    }
  };


  class Set
  {
    public:
      DVector                              Diameter;
      std::vector<complex128>              Index;
      std::vector<std::vector<complex128>> Material;
      DVector                              nMedium;
      bool                                 BoundedIndex;
      std::vector<State>                   States;

      Set(){}
      Set(DVector &Diameter, std::vector<std::vector<complex128>> &Material, DVector &nMedium)
          : Diameter(Diameter), Material(Material), nMedium(nMedium){BoundedIndex=true;}

      Set(DVector &Diameter, std::vector<complex128> &Index, DVector &nMedium)
          : Diameter(Diameter), Index(Index), nMedium(nMedium){BoundedIndex=false;}

      State operator[](size_t idx){return this->States[idx];}

  };
}








namespace CORESHELL
{
  struct State
  {
    double     CoreDiameter,
               ShellDiameter;
    complex128 CoreIndex,
               ShellIndex;
    double     nMedium,
               xCore,
               xShell;
    State(){}
    State(double &CoreDiameter, double &ShellDiameter, complex128 &CoreIndex, complex128& ShellIndex, double &nMedium)
                : CoreDiameter(CoreDiameter), ShellDiameter(ShellDiameter), CoreIndex(CoreIndex), ShellIndex(ShellIndex), nMedium(nMedium){}

    void ApplyMedium()
    {
      this->CoreIndex      /= this->nMedium;
      this->ShellIndex     /= this->nMedium;
      this->CoreDiameter   *= this->nMedium;
      this->ShellDiameter  *= this->nMedium;
    }
  };


  class Set
  {
    public:
      DVector                              CoreDiameter,
                                           ShellDiameter;
      std::vector<complex128>              CoreIndex,
                                           ShellIndex;
      std::vector<std::vector<complex128>> CoreMaterial,
                                           ShellMaterial;
      DVector                              nMedium;
      bool                                 BoundedCore,
                                           BoundedShell;
      std::vector<State>                   States;

      Set(){}

      Set(DVector &CoreDiameter, DVector &ShellDiameter, CVector &CoreIndex, CVector &ShellIndex, DVector &nMedium)
      : CoreDiameter(CoreDiameter), ShellDiameter(ShellDiameter), CoreIndex(CoreIndex), ShellIndex(ShellIndex), nMedium(nMedium)
        { BoundedCore=false; BoundedShell=false; }

      Set(DVector &CoreDiameter, DVector &ShellDiameter, CVector &CoreIndex, std::vector<CVector> &ShellMaterial, DVector &nMedium)
      : CoreDiameter(CoreDiameter), ShellDiameter(ShellDiameter), CoreIndex(CoreIndex), ShellMaterial(ShellMaterial), nMedium(nMedium)
        { BoundedCore=false; BoundedShell=true; }

      Set(DVector &CoreDiameter, DVector &ShellDiameter, std::vector<CVector> &CoreMaterial, CVector &ShellMaterial, DVector &nMedium)
      : CoreDiameter(CoreDiameter), ShellDiameter(ShellDiameter), CoreMaterial(CoreMaterial), ShellIndex(ShellIndex), nMedium(nMedium)
        {BoundedCore=true; BoundedShell=false;}

      Set(DVector &CoreDiameter, DVector &ShellDiameter, std::vector<CVector> &CoreMaterial, std::vector<CVector> &ShellMaterial, DVector &nMedium)
      : CoreDiameter(CoreDiameter), ShellDiameter(ShellDiameter), CoreMaterial(CoreMaterial), ShellMaterial(ShellMaterial), nMedium(nMedium)
        {BoundedCore=true; BoundedShell=true;}

        State operator[](size_t idx){return this->States[idx];}
  };
}





namespace SOURCE
{

    struct State
    {
      double Wavelength;
      CVector Jones;
      double Amplitude;
      size_t WavelengthIdx;
      double k;
      bool Polarized = true;


      State(){}
      State(double  &Wavelength,
            CVector &Jones,
            double &Amplitude)
            : Wavelength(Wavelength),
              Jones(Jones),
              Amplitude(Amplitude)
                  {
                    if (std::isnan( real( Jones[0] ))) Polarized = false;

                    this->k = 2.0 * PI / this->Wavelength;
                  }
    };


    class Set
    {
      public:
        DVector Wavelength;
        std::vector<CVector> Jones;
        DVector Amplitude;
        std::vector<State> States;

        Set(){}
        Set(DVector &Wavelength, std::vector<CVector> &Jones, DVector &Amplitude)
        : Wavelength(Wavelength), Jones(Jones), Amplitude(Amplitude)
        {}

        State operator[](size_t idx){return this->States[idx];}

    };
}








namespace DETECTOR
{
  struct State
  {
    CVector ScalarField;
    double NA, PhiOffset, GammaOffset, Filter;
    bool Coherent, PointCoupling;
    State(){}
    State(CVector &ScalarField,
          double &NA,
          double &PhiOffset,
          double &GammaOffset,
          double &Filter,
          bool   &Coherent,
          bool   &PointCoupling)
          : ScalarField(ScalarField),
          NA(NA),
          PhiOffset(PhiOffset),
          GammaOffset(GammaOffset),
          Filter(Filter),
          Coherent(Coherent),
          PointCoupling(PointCoupling){}
  };



  class Set
  {
    public:
      std::vector<CVector> ScalarField;
      DVector NA, PhiOffset, GammaOffset, Filter;
      bool Coherent, PointCoupling;
      std::vector<State> States;

      Set(){}
      Set(std::vector<CVector> &ScalarField,
          DVector &NA,
          DVector &PhiOffset,
          DVector &GammaOffset,
          DVector &Filter,
          bool    &Coherent,
          bool    &PointCoupling)
      : ScalarField(ScalarField),
        NA(NA),
        PhiOffset(PhiOffset),
        GammaOffset(GammaOffset),
        Filter(Filter),
        Coherent(Coherent),
        PointCoupling(PointCoupling)
      {}

      State operator[](size_t idx){return this->States[idx];}

  };

}



PYBIND11_MODULE(Sets, module)
{
      pybind11::class_<SPHERE::Set>(module, "CppSphereSet")
      .def(pybind11::init<DVector&, std::vector<complex128>&, DVector&>(),
           pybind11::arg("Diameter"),
           pybind11::arg("Index"),
           pybind11::arg("nMedium") )

      .def(pybind11::init<DVector&, std::vector<std::vector<complex128>>&, DVector&>(),
           pybind11::arg("Diameter"),
           pybind11::arg("Material"),
           pybind11::arg("nMedium") )
           ;

      pybind11::class_<CYLINDER::Set>(module, "CppCylinderSet")
      .def(pybind11::init<DVector&, CVector&, DVector&>(),
           pybind11::arg("Diameter"),
           pybind11::arg("Index"),
           pybind11::arg("nMedium")
           )

      .def(pybind11::init<DVector&, std::vector<std::vector<complex128>>&, DVector&>(),
           pybind11::arg("Diameter"),
           pybind11::arg("Material"),
           pybind11::arg("nMedium")
           );


    pybind11::class_<CORESHELL::Set>(module, "CppCoreShellSet")
    .def(pybind11::init<DVector&, DVector&, CVector&, CVector&, DVector&>(),
         pybind11::arg("CoreDiameter"),
         pybind11::arg("ShellDiameter"),
         pybind11::arg("CoreIndex"),
         pybind11::arg("ShellIndex"),
         pybind11::arg("nMedium") )

    .def(pybind11::init<DVector&, DVector&, CVector&, std::vector<CVector>&, DVector&>(),
         pybind11::arg("CoreDiameter"),
         pybind11::arg("ShellDiameter"),
         pybind11::arg("CoreIndex"),
         pybind11::arg("ShellMaterial"),
         pybind11::arg("nMedium"))

    .def(pybind11::init<DVector&, DVector&, std::vector<CVector>&, CVector&, DVector&>(),
         pybind11::arg("CoreDiameter"),
         pybind11::arg("ShellDiameter"),
         pybind11::arg("CoreMaterial"),
         pybind11::arg("ShellIndex"),
         pybind11::arg("nMedium") )

    .def(pybind11::init<DVector&, DVector&, std::vector<CVector>&, std::vector<CVector>&, DVector&>(),
         pybind11::arg("CoreDiameter"),
         pybind11::arg("ShellDiameter"),
         pybind11::arg("CoreMaterial"),
         pybind11::arg("ShellMaterial"),
         pybind11::arg("nMedium") )
         ;

      pybind11::class_<SOURCE::Set>(module, "CppSourceSet")
      .def(pybind11::init<DVector&, std::vector<CVector>&, DVector&>(),
           pybind11::arg("Wavelength"),
           pybind11::arg("Jones"),
           pybind11::arg("Amplitude") );


     pybind11::class_<DETECTOR::Set>(module, "CppDetectorSet")
     .def(pybind11::init<std::vector<CVector>&, DVector&, DVector&, DVector&, DVector&, bool&, bool&>(),
          pybind11::arg("ScalarField"),
          pybind11::arg("NA"),
          pybind11::arg("PhiOffset"),
          pybind11::arg("GammaOffset"),
          pybind11::arg("Filter"),
          pybind11::arg("Coherent"),
          pybind11::arg("PointCoupling")
          );

}
