
from pyface.api import GUI
import numpy, unittest, os, pathlib

from PyMieSim.Tools.Mesh import FibonacciMesh
from PyMieSim.Scatterer import *
from PyMieSim.Source import PlaneWave
from PyMieSim.Detector import Photodiode
from PyMieSim.Experiment import *
from .utils import *
from PyMieSim import Measure, Variables


TestDataPath = pathlib.Path(__file__).parent.parent / "Datas"

PLOTTIME = 2000 # two seconde


class QuantitativeTestCase(unittest.TestCase, BaseStepTest):
    Source = PlaneWave(Wavelength=1e-6, Polarization=0, Amplitude=1)

    sphere = Sphere(Diameter=300e-9, Index=1.4, Source=Source)

    @TestFactory('Validation QSca - CrossSection')
    def step00(self):
        print("\nUnittest: <Quantitative>" + '=' * 100)
        Mesh = FibonacciMesh(MaxAngle=numpy.pi, Sampling=1000, PhiOffset=0, GammaOffset=0)
        val0 = self.sphere.CrossSection(Mesh)
        val1 = self.sphere.Qsca * self.sphere.Area

        if not numpy.isclose(val0, val1, 1e-50):
            raise

    @TestFactory('Validation EnergyFlow - coupling')
    def step01(self):
        detector = Photodiode(Sampling=500, NA=2.0, GammaOffset=0, PhiOffset=0)
        val0 = self.sphere.EnergyFlow(detector.Mesh)
        val1  = detector.Coupling(self.sphere)
        if not numpy.isclose(val0, val1, 1e-10):
            raise

    @TestFactory('Validation scattering cross-section -> Scott Prahl')
    def step02(self):
        Value = Sphere(Diameter=1e-6, Index=1.5, Source=self.Source).Csca
        if not numpy.isclose(Value, 2.7349e-12, 0.001):
            raise

    @TestFactory('Validation g = <cos> -> Scott Prahl')
    def step03(self):
        Value = Sphere(Diameter=1e-6, Index=1.5, Source=self.Source).g
        if not numpy.isclose(Value, 0.72924, 0.001):
            raise

    @TestFactory('Validation extinction cross-section -> Scott Prahl')
    def step04(self):
        Value = Sphere(Diameter=1e-6, Index=1.5 + 1j, Source=self.Source).Cext
        if not numpy.isclose(Value, 2.1422e-12, 0.001):
            raise

    @TestFactory('Validation cylinder with BH figure 8.7')
    def step05(self):
        theoretical = numpy.genfromtxt(f"{TestDataPath}/Figure87BH.csv", delimiter=',')

        Diameter = numpy.geomspace(10e-9, 6e-6, 800)
        Volume = numpy.pi * (Diameter / 2)**2
        scatSet = CylinderSet(Diameter=Diameter, Index=1.55, nMedium=1.)
        sourceSet = SourceSet(Wavelength=632.8e-9, Polarization=[0, 90], Amplitude=1)
        ExpSet = Setup(ScattererSet=scatSet, SourceSet=sourceSet, DetectorSet=None)
        Data = ExpSet.Get(Measure.Csca).Data.squeeze() / Volume * 1e-4 / 100

        GUI.invoke_after(PLOTTIME, Close)
        plt.figure(figsize=(8, 5))
        plt.plot(Diameter, Data[0], 'b--', linewidth=3, label='PyMieSim')
        plt.plot(Diameter, Data[1], 'r-', linewidth=3, label='PyMieSim')

        plt.plot(Diameter, theoretical[0], 'k--', linewidth=1, label='BH 8.8')
        plt.plot(Diameter, theoretical[1], 'k--', linewidth=1, label='BH 8.8')

        plt.xlabel(r'Diameter [$\mu$m]') 
        plt.ylabel('Scattering cross section [Cylinder]')
        plt.grid()
        plt.legend() 
        plt.tight_layout()
        plt.show()

        if not numpy.all(numpy.isclose(Data,theoretical, 1e-9)):
            raise

    @TestFactory('Validation cylinder with BH figure 8.8')
    def step06(self):
        theoretical = numpy.genfromtxt(f"{TestDataPath}/Figure88BH.csv", delimiter=',')

        Diameter = numpy.geomspace(10e-9, 6e-6, 800)
        Volume = numpy.pi * (Diameter / 2)**2
        scatSet = CylinderSet(Diameter=Diameter, Index=1.55, nMedium=1.335)
        sourceSet = SourceSet(Wavelength=632.8e-9, Polarization=[0, 90], Amplitude=1)
        ExpSet = Setup(ScattererSet=scatSet, SourceSet=sourceSet, DetectorSet=None)
        Data = ExpSet.Get(Measure.Csca).Data.squeeze() / Volume * 1e-4 / 100

        GUI.invoke_after(PLOTTIME, Close)
        plt.figure(figsize=(8,5))
        plt.plot(Diameter, Data[0], 'C0-', linewidth=3, label='PyMieSim')
        plt.plot(Diameter, Data[1], 'C1-', linewidth=3, label='PyMieSim')

        plt.plot(Diameter, theoretical[0], 'k--', linewidth=1, label='BH 8.8')
        plt.plot(Diameter, theoretical[1], 'k--', linewidth=1, label='BH 8.8')

        plt.xlabel(r'Diameter [$\mu$m]')
        plt.ylabel('Scattering cross section [Cylinder]')
        plt.grid()
        plt.legend()
        plt.tight_layout()
        plt.show()

        if not numpy.all(numpy.isclose(Data,theoretical, 1e-9)):
            raise

    @TestFactory('Validation PyMieScatt sphere')
    def step07(self):
        theoretical = numpy.genfromtxt(f"{TestDataPath}/PyMieScattQsca.csv", delimiter=',')

        Diameter = numpy.geomspace(10e-9, 6e-6, 800)
        scatSet = SphereSet(Diameter=Diameter, Index=1.4, nMedium=1.)
        sourceSet = SourceSet(Wavelength=632.8e-9, Polarization=[None], Amplitude=1)
        ExpSet = Setup(ScattererSet=scatSet, SourceSet=sourceSet, DetectorSet=None)
        Data = ExpSet.Get(Measure.Qsca).Data.squeeze()

        GUI.invoke_after(PLOTTIME, Close)
        plt.figure(figsize=(8, 5))
        plt.plot(Diameter, Data, 'C1-', linewidth=3, label='PyMieSim')

        plt.plot(Diameter, theoretical, 'k--', linewidth=1, label='PyMieScatt')

        plt.xlabel(r'Diameter [$\mu$m]')
        plt.ylabel('Scattering efficiency [Sphere]')
        plt.grid()
        plt.legend()
        plt.tight_layout()
        plt.show()

        if not numpy.all(numpy.isclose(Data,theoretical, 1e-9)):
            raise

    @TestFactory('Validation PyMieScatt sphere + nMedium')
    def step08(self):
        theoretical = numpy.genfromtxt(f"{TestDataPath}/PyMieScattQscaMedium.csv", delimiter=',')

        Diameter = numpy.geomspace(10e-9, 6e-6, 800)
        scatSet = SphereSet(Diameter=Diameter, Index=1.4, nMedium=1.21)
        sourceSet = SourceSet(Wavelength=632.8e-9, Polarization=[None], Amplitude=1)
        ExpSet = Setup(ScattererSet=scatSet, SourceSet=sourceSet, DetectorSet=None)
        Data = ExpSet.Get(Measure.Qsca).Data.squeeze()

        GUI.invoke_after(PLOTTIME, Close)
        plt.figure(figsize=(8, 5))
        plt.plot(Diameter, Data, 'C1-', linewidth=3, label='PyMieSim')

        plt.plot(Diameter, theoretical, 'k--', linewidth=1, label='PyMieScatt')

        plt.xlabel(r'Diameter [$\mu$m]')
        plt.ylabel('Scattering efficiency [Sphere + nMedium]')
        plt.grid()
        plt.legend()
        plt.tight_layout()
        plt.show()

        if not numpy.all(numpy.isclose(Data,theoretical, 1e-9)):
            raise

    @TestFactory('Validation PyMieScatt CoreShell')
    def step09(self):
        theoretical = numpy.genfromtxt(f"{TestDataPath}/PyMieScattQscaCoreShell.csv", delimiter=',')

        Diameter = numpy.geomspace(10e-9, 500e-9, 400)
        scatSet = CoreShellSet(CoreDiameter = Diameter, ShellDiameter = 600e-9, CoreIndex = 1.4, ShellIndex=1.5, nMedium = 1.0)
        sourceSet = SourceSet(Wavelength = 600e-9, Polarization = [None], Amplitude = 1)
        ExpSet = Setup(ScattererSet = scatSet, SourceSet = sourceSet, DetectorSet  = None)
        Data = ExpSet.Get(Measure.Qsca).Data.squeeze()

        GUI.invoke_after(PLOTTIME, Close)
        plt.figure(figsize=(8,5))
        plt.plot(Diameter, Data,'C1-', linewidth=3, label='PyMieSim' )

        plt.plot( Diameter, theoretical,'k--', linewidth=1, label='PyMieScatt')

        plt.xlabel(r'Diameter [$\mu$m]'); plt.ylabel('Scattering efficiency [CoreShell]')
        plt.grid(); plt.legend(); plt.tight_layout()
        plt.show()

        if not numpy.all(numpy.isclose(Data,theoretical, 1e-9)):
            raise

    @TestFactory('Validation PyMieScatt CoreShell + nMedium')
    def step10(self):
        theoretical = numpy.genfromtxt(f"{TestDataPath}/PyMieScattQscaCoreShellMedium.csv", delimiter=',')

        Diameter = numpy.geomspace(10e-9, 500e-9, 400)
        scatSet = CoreShellSet(CoreDiameter=Diameter, ShellDiameter=600e-9, CoreIndex=1.4, ShellIndex=1.5, nMedium=1.2)
        sourceSet = SourceSet(Wavelength=600e-9, Polarization=[None], Amplitude=1)
        ExpSet = Setup(ScattererSet=scatSet, SourceSet=sourceSet, DetectorSet=None)
        Data = ExpSet.Get(Measure.Qsca).Data.squeeze()

        GUI.invoke_after(PLOTTIME, Close)
        plt.figure(figsize=(8, 5))
        plt.plot(Diameter, Data, 'C1-', linewidth=3, label='PyMieSim')

        plt.plot(Diameter, theoretical, 'k--', linewidth=1, label='PyMieScatt')

        plt.xlabel(r'Diameter [$\mu$m]')
        plt.ylabel('Scattering efficiency [CoreShell + nMedium]')
        plt.grid()
        plt.legend()
        plt.tight_layout()
        plt.show()

        if not numpy.all(numpy.isclose(Data,theoretical, 1e-9)):
            raise

    @TestFactory('Validation cylinder with BH figure 8.10')
    def step11(self):
        theoretical = numpy.genfromtxt(f"{TestDataPath}/Figure810BH.csv", delimiter=',')
        x = theoretical[:, 0]
        y = theoretical[:, 1]

        Source = PlaneWave(Wavelength=470e-9, Polarization=None, Amplitude=1)
        Scat = Cylinder(Diameter=3000e-9, Source=Source, Index=1.00 + 0.07j, nMedium=1.0)
        S1S2 = Scat.S1S2(Phi=x + 90)
        Data = (numpy.abs(S1S2.S1)**2 + numpy.abs(S1S2.S2)**2) * (0.5 / (numpy.pi * Source.k))**(1 / 4)

        error = numpy.abs((Data - y) / y)

        GUI.invoke_after(PLOTTIME, Close)
        plt.figure(figsize=(8, 5))
        plt.plot(S1S2.Phi - 90, Data, 'C1-', linewidth=3, label='PyMieSim')

        plt.plot(x, y, 'k--', linewidth=1, label='B&H [8.10]')

        plt.xlabel('Scattering angle [degree]')
        plt.ylabel('Phase function')
        plt.yscale('log')
        plt.grid()
        plt.legend()
        plt.tight_layout()
        plt.show()

        if not numpy.all(error < 1):
            raise


def suite():
    suite = unittest.TestSuite()
    suite.addTest(QuantitativeTestCase('test_steps'))
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner(failfast=True)
    runner.run(suite())

